import asyncio
import functools


def run_in_executor(func=None, *, context=None, executor=None):
    """
    Decorator to run synchronous function in the given or default executor

    Optionally within the given async context manager (e.g. to limit concurrency
    with `asyncio.Lock` or `asyncio.Semaphore`)
    """

    def decorator(func):
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            # Use dummy Lock() as contextlib.nullcontext for older python versions
            current_context = context or asyncio.Lock()

            loop = asyncio.get_running_loop()
            async with current_context:
                return await loop.run_in_executor(
                    executor, lambda: func(*args, **kwargs)
                )

        return wrapper

    return decorator(func) if func else decorator


run_in_executor_locked = functools.partial(run_in_executor, context=asyncio.Lock())



import copy
import time
import logging
from enum import Enum
from collections import namedtuple
import json
import random
from typing import Any, Dict
import faust
import statistics
import os

CONFIG = dict(
    save_monitoring_logs=None,
    save_monitoring_stats=None,
    SUBPATH_PREFIXES=[],
    default_pipeline_id=None,
    SLIDING_WINDOW_WIDTH=3600,  # = 1hr
)

logger = logging.getLogger(__name__)


def configure(**updates: Dict[str, Any]) -> Dict[str, Any]:
    CONFIG.update(updates)
    return CONFIG


"""
control_point_timing is 2-level structure like
{ path1 : { request_id1: RequestCheckIn( ... ), request_id1: RequestCheckIn( ... ), ... } }
Definitely, the token is a dict key.
path1 is 
"""
control_point_timing = dict()


RequestCheckIn = namedtuple(
    "RequestCheckIn", ["request_id", "timestamp", "elapsed_time"]
)
RequestFootstep = namedtuple(
    "RequestFootstep", ["path", "timestamp", "status", "message"]
)


class Status(str, Enum):
    SUCCESS = "SUCCESS"
    PROGRESS = "PROGRESS"
    ERROR = "ERROR"


class RecordWithMetadata(faust.Record):
    metadata: Any = None

    def __init__(self, **kwargs):
        metadata = kwargs.pop("metadata", None)
        super(faust.Record, self).__init__(**kwargs)
        self.metadata = metadata

    async def trace(self, path, message="", status=Status.PROGRESS):
        if not self.metadata:
            return None
        return await update_metadata(
            path=path,
            status=status,
            metadata=self.metadata,
            message=message,
        )

    async def done(self, path, message="", status=Status.SUCCESS):
        if not self.metadata:
            return
        await update_metadata(
            path=path,
            status=status,
            metadata=self.metadata,
            message=message,
        )
        await post_statistics(self.metadata)


def update_control_point_timings(
    request_id: str, path: str, message_timestamp: float, elapsed_time: float
):
    """
    Extend message timings
    Meaningful paths are
    1) full message path

    2) minimal suffix that starts with one of SUBPATH_ROOTS

    /scraping/start/end/write-to-blob/start/end/translation/start/end/write-to-blob/start/end
    => /write-to-blob/start/end

    """

    # select meaningful suffixes from the path
    positions = list(
        filter(
            lambda x: x >= 0,
            [
                path.find(subpath_prefix, 0)
                for subpath_prefix in CONFIG["SUBPATH_PREFIXES"]
            ],
        )
    )
    if positions:
        _update_subpath_timing(
            request_id, path[max(positions) :], message_timestamp, elapsed_time
        )


def _update_subpath_timing(
    request_id: str, path: str, message_timestamp: float, elapsed_time: float
):
    min_timestamp = time.time() - CONFIG["SLIDING_WINDOW_WIDTH"]
    control_point_timing.setdefault(path, dict())
    control_point_timing[path][request_id] = RequestCheckIn(
        timestamp=message_timestamp, elapsed_time=elapsed_time, request_id=request_id
    )

    old_request_ids = [
        _request_id
        for _request_id, message_check_in in control_point_timing[path].items()
        if message_check_in.timestamp < min_timestamp
    ]
    for old_request_id in old_request_ids:
        control_point_timing[path].pop(old_request_id)


def new_metadata(request_id, pipeline_id=None, prototype={}):
    """
    Creates empty metadata structure
    """
    metadata = copy.deepcopy(prototype)
    metadata["request_id"] = f"{request_id}[{time.time()+0.001*random.random()}]"
    metadata["pipeline_id"] = pipeline_id or CONFIG["default_pipeline_id"]
    metadata["timings"] = list()
    metadata["updated_at"] = time.time()
    return metadata


def copy_metadata(metadata):
    return copy.deepcopy(metadata)


async def update_metadata(*, path, status, metadata, message=""):
    """
    Call this when existing request is processed
    """
    current_time = time.time()

    path_old = get_metadata_path(metadata)
    path_new = os.path.normpath(os.path.join(path_old, path))

    current_footstep = RequestFootstep(
        path=path_new, timestamp=current_time, status=status, message=message
    )

    parent_path = path_new.rsplit("/", 1)[0]
    elapsed_time = 0
    if parent_path:
        for i_footstep in range(len(metadata["timings"]) - 1, -1, -1):
            parent_footstep = metadata["timings"][i_footstep]
            if not isinstance(parent_footstep, RequestFootstep):
                metadata["timings"][i_footstep] = RequestFootstep(*parent_footstep)
                parent_footstep = metadata["timings"][i_footstep]
            if parent_footstep.path == parent_path:
                elapsed_time = current_footstep.timestamp - parent_footstep.timestamp

    metadata["timings"].append(current_footstep)
    metadata["updated_at"] = current_time

    update_control_point_timings(
        metadata["request_id"], path_new, current_time, elapsed_time
    )

    try:
        await CONFIG["save_monitoring_logs"](
            pipeline=metadata["pipeline_id"],
            metadata=json.dumps(
                dict(
                    request_id=metadata["request_id"],
                    pipeline_id=metadata["pipeline_id"],
                    timings=[current_footstep],
                    updated_at=metadata["updated_at"],
                )
            ),
        )
    except Exception as ex:
        logger.error(f"[log_metadata] {ex} {type(ex)}")

    return metadata


def get_metadata_path(metadata):
    if metadata and metadata["timings"]:
        footstep = metadata["timings"][-1]
        if isinstance(footstep, RequestFootstep):
            path_old = footstep.path
        else:
            path_old = RequestFootstep(*footstep).path
    else:
        path_old = ""
    return path_old


async def on_done(*, control_point_id, status, metadata, message=""):
    if not metadata:
        return
    await update_metadata(
        path=control_point_id,
        status=status,
        metadata=metadata,
        message=message,
    )
    await post_statistics(metadata)


async def post_statistics(metadata):
    report_stats = dict()
    min_timestamp = time.time() - CONFIG["SLIDING_WINDOW_WIDTH"]

    for path in control_point_timing:
        mean_elapsed_time = 0

        arrival_rate = len(control_point_timing[path])

        if arrival_rate > 0:
            elapsed_times = [
                message_check_in.elapsed_time
                for message_check_in in control_point_timing[path].values()
                if message_check_in.timestamp >= min_timestamp
            ]
            if elapsed_times:
                mean_elapsed_time = statistics.mean(elapsed_times)
            else:
                mean_elapsed_time = 0
        report_stats[path] = {
            "arrival_rate": arrival_rate,
            "elapsed_time": mean_elapsed_time,
        }

    try:
        await CONFIG["save_monitoring_stats"](
            pipeline=metadata["pipeline_id"],
            updated_at=int(metadata["updated_at"]),
            stats=report_stats,
        )
    except Exception as ex:
        logger.error(f"[on_done] {ex}")



import functools
import logging
import os

from dotenv import load_dotenv
from omegaconf import OmegaConf

logger = logging.getLogger(__name__)


env_path = os.path.abspath(".env")
load_dotenv(dotenv_path=env_path)


@functools.lru_cache
def _load_config(config_dir: str = "./config"):
    # Read default config
    try:
        current_config = OmegaConf.load(f"{config_dir}/default.yml")
    except FileNotFoundError:
        current_config = OmegaConf.create()

    # Read environment-dependent config
    delphai_environment = os.environ.get("DELPHAI_ENVIRONMENT")
    if not delphai_environment:
        raise Exception("DELPHAI_ENVIRONMENT is not defined")

    try:
        environment_config = OmegaConf.load(f"{config_dir}/{delphai_environment}.yml")
        current_config = OmegaConf.merge(current_config, environment_config)
    except FileNotFoundError:
        pass

    OmegaConf.set_readonly(current_config, True)
    return current_config


@functools.lru_cache
def get_config(path: str = "", config_dir: str = "./config"):
    config = _load_config(config_dir=config_dir)
    if path is None:
        return config
    selected = OmegaConf.select(config, path)
    if OmegaConf.is_config(selected):
        return OmegaConf.to_container(selected, resolve=True)
    else:
        return selected



import aio_pika
import aio_pika.connection
import asyncio
import contextvars
import inspect
import logging
import pydantic
import socket
import time

from typing import Optional

from . import errors
from . import metrics
from .connection_manager import get_connection
from .models import Request, Response
from .types import Message, Priority, RequestContext
from .utils import clean_service_name, fix_message_timestamp


logger = logging.getLogger(__name__)

request_context = contextvars.ContextVar("request_context", default=None)


class RpcServer:
    def __init__(self, service_name):
        self._service_name = clean_service_name(service_name)
        self._app_id = f"{self._service_name}@{socket.gethostname()}"

        self._handlers = {}

        self._reset()

        self.bind(self._ping)
        self.bind(self._help)

    def _reset(self):
        self._connection = None
        self._channel = None
        self._exchange = None
        self._queue = None

    def bind(self, handler=None, *, name=None):
        """
        Binds to be exposed handlers (functions) to RPC server instance:

        @server.bind
        def add(*, a: float, b: float) -> float:
            ...

        # or

        def sub(*, a: float, b: float) -> float:
            ...

        server.bind(sub)

        # or

        @server.bind(name="mul")
        def multiply(*, a: float, b: float) -> float:
            ...

        """

        def decorator(handler):
            self._bind_handler(handler=handler, name=name)
            return handler

        return decorator(handler) if handler else decorator

    def _bind_handler(self, *, handler=None, name=None):
        handler_name = name or handler.__name__
        if handler_name in self._handlers:
            raise KeyError(f"Handler {handler_name} already defined")

        if hasattr(handler, "raw_function"):
            # Unwrap `pydantic.validate_call` decorator
            handler = handler.raw_function

        self._validate_handler(handler)

        self._handlers[handler_name] = pydantic.validate_call(
            handler,
            validate_return=True,
        )

    def _validate_handler(self, handler):
        if not inspect.iscoroutinefunction(handler):
            raise TypeError("Handlers must be coroutine functions")

        positional_only = []
        positional_or_keyword = []

        for parameter_name, parameter in inspect.signature(handler).parameters.items():
            if parameter.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                positional_or_keyword.append(parameter_name)

            elif parameter.kind in [
                inspect.Parameter.POSITIONAL_ONLY,
                inspect.Parameter.VAR_POSITIONAL,
            ]:
                positional_only.append(parameter_name)

        if positional_only:
            raise TypeError(
                "{} has positional-only parameters {} that are not supported in {}".format(
                    handler,
                    positional_only,
                    self.__class__,
                )
            )

        if positional_or_keyword:
            logger.warning(
                "%s has positional parameters %s, only keyword parameters are supported in %s",
                handler,
                positional_or_keyword,
                self.__class__,
            )

    async def start(self, connection_string, prefetch_count=1):
        if self._connection:
            raise RuntimeError("Already started")

        connection = self._connection = await get_connection(
            connection_string, self._service_name
        )
        channel = self._channel = await connection.channel()
        await channel.set_qos(prefetch_count=prefetch_count)

        exchange = self._exchange = await channel.declare_exchange(
            name=f"service.{self._service_name}",
            type=aio_pika.ExchangeType.TOPIC,
            durable=True,
        )

        queue = self._queue = await channel.declare_queue(
            name=f"service.{self._service_name}",
            durable=True,
            arguments={
                "x-max-priority": max(Priority),
                "x-dead-letter-exchange": f"service.{self._service_name}.dlx",
            },
        )

        await queue.bind(exchange, "#")
        await queue.consume(fix_message_timestamp(self._on_message))

        logger.info("RPC server is consuming messages from `%s`", queue)

    async def stop(self):
        channel = self._channel
        if channel:
            self._reset()
            await channel.close()

    async def serve_forever(self, connection_string, prefetch_count=1):
        await self.start(connection_string, prefetch_count=prefetch_count)
        try:
            await asyncio.Future()
        finally:
            await self.stop()

    async def _on_message(self, message: Message) -> None:
        consumed_timestamp = time.time()

        logger.debug(
            "[MID:%s] Got `%s` message from `%s` service",
            message.message_id,
            message.type or "[untyped]",
            message.app_id or "unknown",
        )

        metrics_labels = dict(
            exchange=message.exchange,
            routing_key=message.routing_key,
            type=message.type,
            priority=message.priority,
            redelivered=message.redelivered,
            payload_size=message.body_size,
        )

        if message.type != "rpc.request":
            logger.warning(
                "[MID:%s] Unexpected message type: `%s`",
                message.message_id,
                message.type,
            )
            await message.reject()
            metrics.message_consumed(**metrics_labels, error="WRONG_MESSAGE_TYPE")
            return

        if message.reply_to and not message.correlation_id:
            logger.warning("[MID:%s] `correlation_id` is not set", message.message_id)
            metrics.message_consumed(**metrics_labels, error="NO_CORRELATION_ID")
            await message.reject()
            return

        published_timestamp = None
        if message.timestamp:
            published_timestamp = message.timestamp.timestamp()

        deadline = None
        if published_timestamp and message.expiration:
            deadline = published_timestamp + message.expiration

        request_context.set(
            RequestContext(
                deadline=deadline,
                priority=message.priority,
            )
        )

        try:
            response = await asyncio.wait_for(
                self._process_message(message, consumed_timestamp, published_timestamp),
                timeout=(deadline - time.time()) if deadline else None,
            )
        except asyncio.TimeoutError:
            logger.info(
                "[MID:%s] [CID:%s] Execution of `%s` from `%s` service timed out",  # noqa: E501
                message.message_id,
                message.correlation_id,
                message.type,
                message.app_id or "unknown",
            )

            await message.ack()
            metrics.message_consumed(**metrics_labels, error="TIMEOUT")
            return

        if message.reply_to:
            response_message = Message(
                body=response,
                app_id=self._app_id,
                priority=message.priority,
                correlation_id=message.correlation_id,
                expiration=(deadline - time.time()) if deadline else None,
                type="rpc.response",
            )

            await message.channel.basic_publish(
                body=response_message.body,
                routing_key=message.reply_to,
                properties=response_message.properties,
            )
            metrics.message_published(
                exchange="",
                routing_key="",  # random, causes high cardinality metric
                type=response_message.type,
                priority=response_message.priority,
                payload_size=response_message.body_size,
            )

        await message.ack()
        metrics.message_consumed(**metrics_labels)

        logger.debug(
            "[MID:%s] [CID:%s] Handled `%s` from `%s` service, success: %s",  # noqa: E501
            message.message_id,
            message.correlation_id,
            message.type,
            message.app_id or "unknown",
            response.error is None,
        )

    @Response.wrap_errors
    async def _process_message(
        self,
        message: Message,
        consumed_timestamp: float,
        published_timestamp: Optional[float],
    ) -> Response:
        request = Request.model_validate_message(message)

        timings = request.timings
        queued_for = None
        if published_timestamp is not None:
            timings.append(("queue.published", published_timestamp))
            queued_for = consumed_timestamp - published_timestamp

        timings.append((f"queue.consumed by {self._app_id}", consumed_timestamp))

        with metrics.server_requests_in_progress.labels(
            method=request.method_name
        ).track_inprogress():
            elapsed = -time.perf_counter()
            response = await self._process_request(request)
            elapsed += time.perf_counter()
        timings.append(("execution.completed", consumed_timestamp + elapsed))

        response.context = request.context
        response.timings = timings

        metrics.server_request_processed(
            priority=message.priority,
            method=request.method_name,
            error=response.error,
            queued_for=queued_for,
            elapsed=elapsed,
        )

        logger.info(
            "[MID:%s] Processed `%s` from `%s` service to method `%s`. In queue: %ims, execution: %ims, success: %s%s",
            message.message_id,
            message.type,
            message.app_id or "unknown",
            request.method_name,
            max(queued_for * 1000, 0),
            elapsed * 1000,
            response.error is None,
            (f", error: {response.error.message}" if response.error else ""),
        )

        return response

    @Response.wrap_errors
    async def _process_request(self, request) -> Response:
        handler = self._handlers.get(request.method_name)
        if handler is None:
            raise errors.UnknownMethodError(request.method_name)

        try:
            result = await handler(**request.arguments)
        except Exception as error:
            raise errors.ExecutionError(repr(error))

        if isinstance(result, pydantic.BaseModel):
            result = result.model_dump()

        return Response(result=result)

    async def _ping(self) -> None:
        return None

    async def _help(self):
        """
        Returns methods list
        """
        return {
            "methods": [
                {
                    "method_name": method_name,
                    "signature": f"{method_name}{inspect.signature(handler)}",
                    "description": handler.__doc__,
                }
                for method_name, handler in self._handlers.items()
            ]
        }



import logging as _logging

import coloredlogs

from delphai_utils.config import get_config


LOGGING_LEVEL = "INFO"

LOGGING_FORMAT = "[%(asctime)s.%(msecs)03d] [%(levelname)s] [%(name)s] %(message)s"

LOGGERS_RESET = [
    "uvicorn",
    "uvicorn.error",
    "uvicorn.access",
]

LOGGERS_LEVELS = {
    "aiocache": "ERROR",
    "aiohttp": "ERROR",
    "aiokafka": "ERROR",
    "azure": "ERROR",
    "elasticsearch": "ERROR",
    "faust": "ERROR",
    "httpx": "WARNING",
    "mode": "ERROR",
    "urllib3": "ERROR",
}


def configure_logging(*, loggers_levels={}):
    logging_config = get_config("logging") or {}

    coloredlogs.install(
        level=_logging.NOTSET, fmt=logging_config.get("format", LOGGING_FORMAT)
    )

    for logger_name in LOGGERS_RESET:
        logger_object = _logging.getLogger(logger_name)
        logger_object.propagate = True
        logger_object.setLevel(_logging.NOTSET)
        logger_object.handlers.clear()

    _logging.root.setLevel(logging_config.get("level", LOGGING_LEVEL))

    loggers_levels = dict(LOGGERS_LEVELS, **(loggers_levels or {}))
    for logger_name, logger_level in loggers_levels.items():
        _logging.getLogger(logger_name).setLevel(logger_level)

    for logger_name, logger_config in logging_config.get("loggers", {}).items():
        if "level" in logger_config:
            _logging.getLogger(logger_name).setLevel(logger_config["level"])



class RpcError(Exception):
    pass


class TemporaryError(RpcError):
    pass


class FinalError(RpcError):
    pass


class UnknownError(FinalError):
    pass


class UnhandledError(FinalError):
    pass


class ParsingError(FinalError):
    pass


class UnknownServiceError(FinalError):
    pass


class UnknownMethodError(FinalError):
    pass


class ExecutionError(FinalError):
    pass



import re
from typing import Set


def get_bow(sentence: str) -> Set[str]:
    return {x for x in re.split(r"\W+", sentence.lower()) if x}


def get_trigrams(sentence: str) -> Set[str]:
    if not sentence:
        return set()
    clear_sentence = " ".join([x for x in re.split(r"\W+", sentence.lower()) if x])
    return {clear_sentence[i - 3 : i] for i in range(3, len(clear_sentence) + 1)}


def similarity(s1: str, s2: str) -> float:
    s1_set: Set[str] = get_trigrams(s1)
    s2_set: Set[str] = get_trigrams(s2)
    return len(s1_set & s2_set) / max(1, len(s1_set | s2_set))



import aio_pika.message
import functools
import logging
import msgpack
import pydantic
import zlib

from typing import Dict, Any, List, Optional, Tuple

from . import errors


logger = logging.getLogger(__name__)


class BaseModel(pydantic.BaseModel):
    @classmethod
    def model_validate_message(cls, message: aio_pika.message.Message):
        body = message.body
        if message.content_encoding == "deflate":
            try:
                body = zlib.decompress(body)
            except Exception as error:
                raise errors.ParsingError(f"Message decompression failed: `{error!r}`")

        elif message.content_encoding:
            raise errors.ParsingError(
                f"Unknown content_encoding: `{message.content_encoding}`"
            )

        if message.content_type != "application/msgpack":
            raise errors.ParsingError(
                f"Got a message with unknown content type: {message.content_type}"
            )

        try:
            return cls.model_validate(msgpack.loads(body))
        except ValueError as error:
            raise errors.ParsingError(f"Message deserialization failed: `{error!r}`")

    def model_dump_msgpack(self, **kwargs) -> bytes:
        kwargs.setdefault("exclude_defaults", True)
        return msgpack.dumps(self.model_dump(**kwargs))


class Request(BaseModel):
    method_name: str
    arguments: Dict[str, Any] = {}
    context: Optional[Any] = None
    timings: List[Tuple[str, float]] = []


class ResponseError(BaseModel):
    type: str
    message: Optional[str] = None


class Response(BaseModel):
    result: Optional[Any] = None
    error: Optional[ResponseError] = None
    context: Optional[Any] = None
    timings: List[Tuple[str, float]] = []

    @classmethod
    def wrap_errors(cls, func):
        @functools.wraps(func)
        async def inner(*args, **kwargs):
            try:
                response = await func(*args, **kwargs)
                if not isinstance(response, cls):
                    raise TypeError(f"Incorrect response type, got: {type(response)}")

            except errors.RpcError as error:
                response = cls(
                    error={"type": type(error).__name__, "message": error.args[0]}
                )

            except Exception as error:
                logger.exception("Unhandled error")
                response = cls(error={"type": "UnhandledError", "message": repr(error)})

            return response

        return inner



import datetime
import functools
import re


def fix_message_timestamp(func):
    @functools.wraps(func)
    def inner(message):
        # Fix `pamqp` naive timestamp
        if message.timestamp:
            message.timestamp = message.timestamp.replace(tzinfo=datetime.timezone.utc)

        return func(message)

    return inner


def clean_service_name(service_name):
    return re.sub("[^a-z0-9-]+", "-", service_name.strip().lower())



COUNTRY_LEGAL_ENTITY_TYPES_MAPPING = {
    "Albania": ["sh.a.", "sh.p.k."],
    "Argentina": [
        "s.a.",
        "s.r.l.",
        "s.c.p.a",
        "scpa",
        "s.c.e i.",
        "s.e.",
        "s.g.r",
        "soc.col.",
        "s.a.s.",
        "s.a.u.",
        "scs",
        "s.c.e l.",
    ],
    "Australia": ["nl", "pty. ltd.", "pty ltd", "llp", "ilp", "inc.", "trust"],
    "Austria": [
        "e.u.",
        "stg",
        "gesbr",
        "a.g.",
        "ag",
        "og",
        "kg",
        "gen",
        "privatstiftung",
        "gmbh",
    ],
    "Belarus": ["aat", "3at", "sat", "pup", "taa"],
    "Belgium": [
        "esv",
        "vzw",
        "vof",
        "snc",
        "comm.v",
        "scs",
        "bvba",
        "sprl",
        "cvba",
        "cvoa",
        "sca",
        "sep",
        "gie",
        "asbl",
        "comm. va",
        "bv",
        "srl",
        "ebvba",
        "sprlu",
        "nv",
        "sa",
        "scrl",
        "scri",
    ],
    "Bosnia / Herzegovina": [
        "d.d.",
        "a.d.",
        "d.n.o.",
        "d.o.o.",
        "k.v.",
        "s.p.",
        "k.d.",
    ],
    "Brazil": ["ltda", "s.a.", "pllc", "ad", "adsitz", "ead", "et", "kd", "kda", "sd"],
    "Brunei": ["Sdn. Bhd.", "bhd."],
    "Bulgaria": ["ad", "adsits", "ead", "et", "kd", "kda", "sd", "eood", "ood"],
    "Cambodia": ["gp", "sm pte ltd.", "pte ltd.", "plc ltd.", "peec", "sp"],
    "Canada": ["gp", "lp", "sp"],
    "Chile": [
        "eirl",
        "s.a.",
        "sgr",
        "s.g.r.",
        "ltda",
        "s.p.a.",
        "sa",
        "s. en c.",
        "ltda.",
    ],
    "China": [
        "ltd.",
        "plc",
        "有限责任公司",
        "股份有限公司",
        "个人独资企业",
        "合伙制企业",
    ],
    "Columbia": ["s.a.", "e.u.", "s.a.s.", "suc. de descendants", "sca", "ltda"],
    "Costa Rica": ["s.a.", "s.r.l.", "ltda."],
    "Croatia": [
        "d.d.",
        "d.o.o.",
        "obrt",
        "j.d.o.o.",
        "j.t.d.",
        "k.d.",
        "giu",
        "zadruga",
        "obrt",
        "opg",
    ],
    "Cyprus": ["ltd", "plc", "coop", "sp", "sf", "publaw", "pf", "oth"],
    "Czech Republic": [
        "a.s.",
        "akc. spol.",
        "s.r.o.",
        "spol. s r.o.",
        "v.o.s.",
        "obch. spol.",
        "a spol.",
        "k.s.",
        "kom. spol.",
        "kom. spol.",
        "živnost",
        "s.p.",
        "příspěvková organizace",
        "z.s.",
    ],
    "Denmark": [
        "i/s",
        "a/s",
        "k/s",
        "p/s",
        "amba",
        "a.m.b.a.",
        "fmba",
        "f.m.b.a.",
        "smba",
        "s.m.b.a.",
        "g/s",
        "forening",
        "ivs",
        "aps",
    ],
    "Dominican Republic": [
        "c. por a.",
        "cxa",
        "s.a.",
        "s.a.s.",
        "srl.",
        "srl",
        "eirl.",
        "sa",
        "sas",
    ],
    "Ecuador": ["s.a.", "c.a.", "sa", "ep", "s.a.s.", "cia. ltda."],
    "Egypt": ["sae", "llc", "llp"],
    "Ethiopia": ["plc", "sc", "cs", "pe", "pc"],
    "Estonia": ["fie", "UÜ", "TÜ", "OÜ", "as", "MTÜ"],
    "Finland": [
        "t:mi",
        "tmi",
        "as oy",
        "as.oy",
        "ay",
        "ky",
        "oy",
        "oyj",
        "ok",
        "osk",
        "T:mi",
        "ry",
        "rp",
        "rs",
    ],
    "France": [
        "sicav",
        "sarl",
        "sogepa",
        "ei",
        "eurl",
        "sasu",
        "fcp",
        "gie",
        "sep",
        "snc",
        "scs",
        "sca",
        "scop",
        "sem",
        "sas",
        "ei",
        "fcp",
        "sppl",
        "sci",
        "sarl",
        "eurl",
        "sa",
        "scop",
        "sem",
        "sas",
        "sasu",
    ],
    "Germany": [
        "gmbh & co. kg",
        "e.g.",
        "e.v.",
        "gbr",
        "ohg",
        "partg",
        "kgaa",
        "gmbh",
        "g.m.b.h.",
        "ag",
        "mbh & co. kg",
        "kg",
        "AG & Co. KG",
        "SE & Co. KG",
        "GmbH & Co. OHG",
        "Limited & Co. KG",
        "PLC & Co. KG",
        "ApS & Co. KG",
        "LLC & Co. KG",
        "PartGmbBH",
        "PartG",
        "e.V.",
        "gGmbH",
        "UG",
        "e.G.",
    ],
    "Ghana": ["llc", "clg", "ulc"],
    "Greece": [
        "a.e.",
        "ae",
        "e.e.",
        "ee",
        "epe",
        "e.p.e.",
        "mepe",
        "m.e.p.e.",
        "o.e.",
        "oe",
        "ovee",
        "o.v.e.e.",
        "A.V.E.E.",
        "I.K.E.",
    ],
    "Guatemala": [
        "s.a.",
        "sa",
        "y Cía. S. C.",
        "y Cía. S. en C.",
        "Cía. Ltda.",
        "Cía. S. C. A",
    ],
    "Haiti": ["sa"],
    "Hong Kong": ["ltd", "unltd", "ultd", "limited"],
    "Hungary": [
        "e.v.",
        "e.c.",
        "bt.",
        "kft.",
        "kht.",
        "kkt.",
        "k.v.",
        "zrt.",
        "nyrt",
        "ev",
        "ec",
        "rt.",
        "szöv.",
    ],
    "Iceland": [
        "ehf.",
        "hf.",
        "ohf.",
        "s.f.",
        "ses.",
        "ehf. einkahlutafélag",
        "einstaklingsfyrirtæki",
        "hf. hlutafélag",
        "ohf. opinbert hlutafélag",
        "slf. samlagsfélag",
        "svf. samvinnufélag",
        "sf. sameignarfélag",
        "ses. sjálfseignarstofnun",
    ],
    "India": ["pvt. ltd.", "ltd.", "psu", "pse", "gcmmf"],
    "Indonesia": ["ud", "fa", "pt", "vof/fa", "mts", "cv", "pt tbk"],
    "Iran": [
        "شرکت سهامی عام",
        "شرکت سهامی خاص",
        "شرکت با مسئولیت محدود",
        "شرکت مختلط غیر سهامی",
        "شرکت مختلط سهامی",
        "شرکت مختلط سهامی",
        "شرکت تضامنی",
        "شرکت نسبی",
        "شرکت تعاونی تولید و مصرف",
    ],
    "Ireland": ["cpt", "teo", "ucits", "dac", "spv"],
    "Israel": ["b.m.", "bm", "ltd", "limited", "cc"],
    "Italy": [
        "s.n.c.",
        "s.a.s.",
        "s.p.a.",
        "s.a.p.a.",
        "s.r.l.",
        "s.c.r.l.",
        "s.s.",
        "s.c.p.a.",
        "s.c.a.r.l.",
        "s.r.l.s.",
    ],
    "Japan": [
        "k.k.",
        "co., ltd.",
        "company, limited",
        "g.k.",
        "y.k.",
        "gmk",
        "gsk",
        "nk",
        "tk",
        "株式会社",
        "合同会社",
        "有限会社",
        "合名会社",
        "合資会社",
        "任意組合",
        "匿名組合",
        "投資事業有限責任組合",
        "有限責任事業組合",
    ],
    "Kazakhstan": [
        "ao",
        "too",
        "tdo",
        "gp",
        "kt",
        "oo",
        "pt",
        "ptk",
        "prk",
        "po",
        "uchr",
    ],
    "Kosovo": ["b.i.", "o.p.", "sh.k.m.", "sh.p.k.", "j.s.c.", "sh.a."],
    "Latvia": ["as", "sia", "ik", "ps", "ks", "ĀKF", "bo", "vsia"],
    "Liechtenstein": ["ag", "gmbh"],
    "Lebanon": ["sal", "s.a.l.", "s.a.r.l.", "s.n.c.", "s.c.s.", "rr"],
    "Lithuania": ["uab", "ab", "ij", "mb", "VšĮ", "IĮ", "TŪB", "KŪB", "BĮ", "pp"],
    "Luxembourg": ["s.a.", "s.a.r.l.", "secs", "SARL-S", "scs", "scsp", "sca"],
    "Macedonia": ["d.o.o.", "d.o.o.e.l", "k.d.a.", "j.t.d.", "a.d.", "k.d."],
    "Malta": ["ltd", "plc"],
    "Malaysia": ["bhd.", "sdn. bhd.", "plt"],
    "Mexico": [
        "s.a.",
        "s. de. r.l.",
        "s. en c.",
        "s.a.b.",
        "s.a.p.i.",
        "S. en C. por A.",
    ],
    "Monaco": ["sam", "sarl", "snc", "scs", "sca"],
    "Mongolia": ["xk", "xxk"],
    "Montenegro": ["o.d.", "k.d.", "a.d.", "d.o.o."],
    "Morocco": ["s.a.", "s.a.s.", "s.a.r.l.", "g.i.e.", "s.n.c.", "s.c.s.", "s.c.a."],
    "Netherlands": [
        "v.o.f.",
        "c.v.",
        "b.v.",
        "n.v.",
        "w.a.",
        "b.a.",
        "u.a.",
        "vve",
        "mts",
        "vof",
        "cv",
        "bv",
        "nv",
    ],
    "New Zealand": ["tapui", "ltd", "limited"],
    "Nicaragua": ["s.a.", "cia", "cia. ltda.", "s.c.s.", "s.c.a."],
    "Nigeria": ["gte.", "plc", "ltd.", "ultd.", "ltd./gte.", "llp"],
    "Norway": [
        "asa",
        "as",
        "ans",
        "ba",
        "bl",
        "da",
        "etat",
        "fkf",
        "hf",
        "iks",
        "kf",
        "ks",
        "nuf",
        "rhf",
        "sf",
        "ub",
    ],
    "North Macedonia": ["a.d.", "d.o.o.", "d.o.o.e.l.", "k.d.", "k.d.a.", "j.t.d."],
    "Oman": ["saog", "saoc"],
    "Pakistan": ["ltd.", "pvt. ltd.", "ltd", "limited"],
    "Peru": ["sa", "s.a.", "s.a.a.", "s.a.c."],
    "Philippines": [
        "coop.",
        "co.",
        "opc",
        "corp.",
        "corp",
        "ent.",
        "inc.",
        "inc",
        "llc",
        "l.l.c.",
        "ltd.",
        "ltd. co.",
        "llp",
        "cia",
        "sa",
        "s.a.",
    ],
    "Poland": [
        "p.p.",
        "s.k.a.",
        "sp.j.",
        "sp.k.",
        "sp.p.",
        "sp. z o.o.",
        "s.c.",
        "s.a.",
        "p.s.a.",
    ],
    "Portugal": ["lda.", "crl", "s.a.", "s.f.", "sgps", "sua", "sgoic", "sic"],
    "Romania": ["s.c.a.", "s.c.s.", "s.n.c.", "s.r.l.", "o.n.g.", "s.a.", "pfa"],
    "Russia": [
        "ooo",
        "oao",
        "zao",
        "3ao",
        "пао",
        "оао",
        "ооо",
        "ngo",
        "npo",
        "ip",
        "gp",
        "gup",
        "mp",
        "mup",
        "ao",
    ],
    "Saudi Arabia": [
        "شركة ذات مسئولية محدودة",
        "شركة مساهمة",
        "شركة تضامن",
        "شركة التوصية البسيطة",
        "شركة أجنبية ",
        "مؤسسة فردية",
    ],
    "Serbia": ["d.o.o.", "a.d.", "k.d.", "o.d.", "d.d.", "j.p."],
    "Singapore": [
        "bhd",
        "pte ltd",
        "sdn bhd",
        "llp",
        "l.l.p.",
        "ltd.",
        "pte",
        "pte. ltd.",
        "llp",
    ],
    "Slovenia": ["d.d.", "d.o.o.", "d.n.o.", "k.d.", "s.p."],
    "Slovakia": [
        "a.s.",
        "akc. spol.",
        "s.r.o.",
        "spol. s r.o.",
        "k.s.",
        "kom. spol.",
        "v.o.s.",
        "a spol.",
        "družstvo",
        "živnosť",
    ],
    "South Korea": [
        "합명회사",
        "합자회사",
        "유한책임회사",
        "주식회사;",
        "유한회사",
        "협동조합",
        "협동조합연합회",
    ],
    "South Africa": ["(pty) ltd", "ltd.", "cc"],
    "Spain": [
        "s.a.",
        "s.a.d.",
        "s.l.",
        "s.l.l.",
        "s.l.n.e.",
        "s.c.",
        "s.cra",
        "s.coop",
        "sal",
        "sccl",
        "c.b.",
        "s.c.p.",
        "S.Com. p. A.",
        "S.Com p.A.P.",
        "ute",
        "scr",
        "fcr",
        "sicc",
        "ficc",
        "sgeic",
    ],
    "Sweden": ["ab", "hb", "kb", "ek"],
    "Switzerland": [
        "ab",
        "sa",
        "gmbh",
        "g.m.b.h.",
        "sarl",
        "sagl",
        "ag",
        "sicaf",
        "wig",
        "ri",
        "eg",
        "koig",
        "kg",
        "komag",
    ],
    "Tunisia": ["sa", "sarl", "suarl"],
    "Thailand": [
        "บริษัทมหาชนจำกัด",
        "บริษัทเอกชนจำกัด",
        "ห้างหุ้นส่วนจำกัด",
        "ห้างหุ้นส่วนสามัญนิติบุคคล",
    ],
    "Taiwan": ["無限公司", "有限公司", "兩合公司", "股份有限公司"],
    "Turkey": [
        "koop.",
        "Adi Şirket",
        "Koll. Şti.",
        "Kom. Şti",
        "A.Ş. / A.O.",
        "Ltd. Şti. / L.Ş. / L.O.",
        "HAAO",
        "koop.",
        "Şahıs Firması",
        "Şb.",
    ],
    "Uganda": ["llc", "plc"],
    "Ukraine": [
        "dat",
        "fop",
        "kt",
        "pt",
        "tdv",
        "tov",
        "pp",
        "vat",
        "zat",
        "at",
        "pat",
        "prat",
        "at",
        "ДАТ",
        "ФОП",
        "ПT",
        "ТДВ",
        "TOB",
        "ПП",
        "ПАТ",
        "рАТ",
    ],
    "United Kingdom": [
        "plc.",
        "plc",
        "cic",
        "cio",
        "l.l.p.",
        "llp",
        "l.p.",
        "lp",
        "ltd.",
        "ltd",
        "limited",
        "slp",
        "ultd.",
        "unltd.",
    ],
    "United Arab Emirates": ["fze", "fzc", "llc"],
    "United States": [
        "llc",
        "inc.",
        "corporation",
        "incorporated",
        "company",
        "limited",
        "corp.",
        "inc.",
        "inc",
        "llp",
        "l.l.p.",
        "pllc",
        "and company",
        "& company",
        "inc",
        "inc.",
        "corp.",
        "corp",
        "ltd.",
        "ltd",
        "& co.",
        "& co",
        "co.",
        "co",
        "lp",
        "smllc",
        "pac",
        "fsb",
        "na",
        "NT&SA",
        "ncua",
        "lp",
        "lllp",
        "lc",
        "llc",
        "pllc",
        "l3c",
        "pbc",
        "pc",
        "p.c.",
        "co-op",
        "coop",
        "cp",
        "dba",
    ],
    "Uruguay": ["s.a.", "s.a.s.", "s.r.l.", "s.c.i.", "s.c."],
    "Uzbekistan": ["mchj", "qmj", "aj", "oaj", "yoaj", "xk", "xt", "ok", "uk", "qk"],
    "Venezuela": ["s.a.", "s.r.l.", "s.c.s.", "soc. col."],
    "Vietnam": [
        "Cty TNHH",
        "Cty TNHH MTV",
        "Cty CP",
        "DNNN",
        "DNTN",
        "DTNN",
        "htx",
        "Chi nhánh",
        "Nhóm Cty",
    ],
    "Zambia": ["ulc", "llc"],
    "Zimbabwe": ["plc", "pbc"],
}
